# Generated by Django 2.1.3 on 2019-01-21 16:49

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [("mds", "0002_area_color")]

    operations = [
        migrations.AddField(
            model_name="eventrecord",
            name="timestamp",
            field=models.DateTimeField(
                db_index=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="eventrecord",
            name="saved_at",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterUniqueTogether(
            name="eventrecord", unique_together={("device", "timestamp")}
        ),
    ]



"""
Pulling data for registered providers

This is the opposite of provider pushing their data to the agency API.
"""
import datetime
import enum
import logging
import urllib.parse
import uuid

from django.contrib.gis import geos
from django.db import transaction
from django.utils import timezone

import requests
from retrying import retry

from mds import db_helpers
from mds import enums
from mds import models
from mds import utils
from mds.provider_mapping import PROVIDER_REASON_TO_AGENCY_EVENT
from .oauth2_store import OAuth2Store
from .translation import translate_data
from .settings import PROVIDER_POLLER_LIMIT_DAYS


ACCEPTED_MDS_VERSIONS = ["0.2", "0.3"]
POLLING_CURSOR_TO_QUERY_PARAM = enum.Enum(
    "polling_cursor field to the query parameter",
    [
        ("start_recorded", "start_recorded"),
        ("start_time", "start_time"),
        ("total_events", "skip"),
    ],
)

logger = logging.getLogger(__name__)


def get_date_param(last_event):
    if last_event:
        return utils.to_mds_timestamp(last_event)
    elif PROVIDER_POLLER_LIMIT_DAYS:
        delta = timezone.now() - datetime.timedelta(PROVIDER_POLLER_LIMIT_DAYS)
        logger.debug(f"No last_event on provider, starting from: {str(delta)}")
        return utils.to_mds_timestamp(delta)


class StatusChangesPoller:
    """Poll all providers to fetch their latest telemetry data."""

    def __init__(self, provider, cursor=None, from_cursor=None, to_cursor=None):
        self.provider = provider
        self.overriden_cursor = cursor
        self.from_cursor = from_cursor
        self.to_cursor = to_cursor
        self.oauth2_store = OAuth2Store(provider)
        # While we poll a given provider, it may aggegrate data from several providers
        # (these are sets of UUIDs, not str)
        self.provider_uids = set(models.Provider.objects.values_list("pk", flat=True))
        self.device_uids = set(models.Device.objects.values_list("pk", flat=True))

    def poll(self):
        if not self.provider.base_api_url:
            logger.debug("Provider %s has no URL, skipping.", self.provider.name)
            return
        logger.debug(f"Polling {self.provider.name}")
        self._poll_status_changes()

    def _poll_status_changes(self):
        next_url = urllib.parse.urljoin(self.provider.base_api_url, "status_changes")
        if self.provider.api_configuration.get("trailing_slash"):
            next_url += "/"

        params = {}
        # Start where we left, it's all based on providers sorting by start_time
        # (but we would miss telemetries older than start_time saved after we polled).
        # For those that support it, use the recorded time field or equivalent.
        polling_cursor = self.overriden_cursor or self.provider.api_configuration.get(
            "polling_cursor", POLLING_CURSOR_TO_QUERY_PARAM.start_time.name
        )
        logger.info(
            f"Starting polling using the field: {polling_cursor}. Current state:\n"
            + (
                (
                    f"\tLast event_time: {str(self.provider.last_event_time_polled)},\n"
                    + f"\tLast recorded: {str(self.provider.last_recorded_polled)}\n"
                    + f"\tLast skip: {self.provider.last_skip_polled}."
                )
                if not self.overriden_cursor
                else f"\tCursor value: {self.from_cursor}."
            )
        )

        # skip
        if polling_cursor == POLLING_CURSOR_TO_QUERY_PARAM.total_events.name:
            params[POLLING_CURSOR_TO_QUERY_PARAM.total_events.value] = (
                self.from_cursor or self.provider.last_skip_polled
            )
        # start_recorded
        elif polling_cursor == POLLING_CURSOR_TO_QUERY_PARAM.start_recorded.name:
            params[POLLING_CURSOR_TO_QUERY_PARAM.start_recorded.value] = get_date_param(
                self.from_cursor or self.provider.last_recorded_polled
            )
        # start_time
        elif polling_cursor == POLLING_CURSOR_TO_QUERY_PARAM.start_time.name:
            params[POLLING_CURSOR_TO_QUERY_PARAM.start_time.value] = get_date_param(
                self.from_cursor or self.provider.last_event_time_polled
            )

        # Provider-specific params to optimise polling
        try:
            params.update(self.provider.api_configuration["status_changes_params"])
        except KeyError:
            pass

        if params:
            next_url = "%s?%s" % (next_url, urllib.parse.urlencode(params))

        skip_polled = (
            self.from_cursor
            if self.from_cursor
            and self.overriden_cursor == POLLING_CURSOR_TO_QUERY_PARAM.total_events.name
            else None
        )

        # Pagination
        while next_url:
            body = self._get_body(next_url)
            # Translate older versions of data
            translated_data = translate_data(body["data"], body["version"])
            status_changes = translated_data["status_changes"]
            if not status_changes:
                break

            next_url = body.get("links", {}).get("next")

            # A transaction for each "page" of data
            with transaction.atomic():
                # We get the maximum of the recorded and event_types
                # from the status changes
                event_time_polled, recorded_polled = self._process_status_changes(
                    status_changes
                )
                if self.overriden_cursor:
                    if (
                        polling_cursor
                        == POLLING_CURSOR_TO_QUERY_PARAM.total_events.name
                    ):
                        skip_polled += len(status_changes)
                        if skip_polled >= self.to_cursor:
                            break
                    elif (
                        polling_cursor
                        == POLLING_CURSOR_TO_QUERY_PARAM.start_recorded.name
                    ):
                        if recorded_polled >= self.to_cursor:
                            break
                    elif (
                        polling_cursor == POLLING_CURSOR_TO_QUERY_PARAM.start_time.name
                    ):
                        if event_time_polled >= self.to_cursor:
                            break
                else:
                    # We get the new skip from the number of status changes
                    skip_polled = (
                        self.provider.last_skip_polled + len(status_changes)
                        if self.provider.last_skip_polled
                        else len(status_changes)
                    )
                    self.provider.last_event_time_polled = event_time_polled
                    self.provider.last_recorded_polled = recorded_polled
                    self.provider.last_skip_polled = skip_polled
                    self.provider.save(
                        update_fields=[
                            "last_event_time_polled",
                            "last_recorded_polled",
                            "last_skip_polled",
                        ]
                    )

                logger.info(
                    f"Polled page using cursor: {polling_cursor}. New state:\n"
                    + f"\tLast event_time: {str(event_time_polled)},\n"
                    + f"\tLast recorded: {str(recorded_polled)}\n"
                    + f"\tLast skip: {skip_polled}."
                )

    @retry(stop_max_attempt_number=2)
    def _get_body(self, url):
        authentication_type = self.provider.api_authentication.get("type")
        if authentication_type in (None, "", "none"):
            client = requests
        elif authentication_type == "oauth2":
            client = self.oauth2_store.get_client()
        else:
            raise NotImplementedError

        headers = {}
        accepted_content_types = ["application/json"]
        # Add versioning header (some providers may choke on this)
        for version in ACCEPTED_MDS_VERSIONS:
            accepted_content_types.append(
                "application/vnd.mds.provider+json;version=%s" % version
            )
        headers["Accept"] = ", ".join(accepted_content_types)

        logger.debug("Polling provider on URL %s", url)
        response = client.get(url, timeout=30, headers=headers)
        # Token may be expired sooner than expected, retry in one minute
        if response.status_code in (401, 403):
            self.oauth2_store.flush_token()
        response.raise_for_status()

        # Servers should send what version of the API was served
        # but we rather trust the versioning in the body
        content_type = response.headers.get(
            "Content-Type", "application/json; charset=UTF-8"
        )
        if content_type.startswith("application/json"):
            version = "unspecified"
        else:
            # Let it raise if malformed
            content_type, version = content_type.split(";")
            _, version = version.split("=")
        logger.debug(
            "Provider %s pretends to accept version %s", self.provider.name, version
        )

        body = response.json()
        return body

    def _process_status_changes(self, status_changes):
        logger.debug("Processing...")

        # accept timestamp as a string instead of an integer
        status_changes = self._validate_event_times(status_changes)
        if not status_changes:
            # Data so bad there is no or nothing but invalid event times
            logger.exception(
                "No valid event_time found in status_changes series: %s", status_changes
            )
            # How can we prevent from asking them again next time?
            if (
                self.provider.last_event_time_polled
                and self.provider.last_recorded_polled
            ):
                ms = datetime.timedelta(milliseconds=1)
                return (
                    self.provider.last_event_time_polled + ms,
                    self.provider.last_recorded_polled + ms,
                )

            # The provider really doesn't help!
            return timezone.now(), timezone.now()

        last_event_time_polled = utils.from_mds_timestamp(
            max(status_change["event_time"] for status_change in status_changes)
        )
        last_recorded_polled = utils.from_mds_timestamp(
            max(status_change.get("recorded") or 0 for status_change in status_changes)
        )

        status_changes = self._validate_status_changes(status_changes)
        if not status_changes:
            # None were valid, we won't ask that series again
            # (provided status changes are ordered by event_time ascending)
            return last_event_time_polled, last_recorded_polled

        self._create_missing_providers(status_changes)
        self._create_missing_devices(status_changes)
        self._create_event_records(status_changes)

        return last_event_time_polled, last_recorded_polled

    def _validate_event_times(self, status_changes):
        """I need this one done before validating the rest of the data."""
        validated_status_changes = []

        for status_change in status_changes:
            try:
                status_change["event_time"] = int(status_change["event_time"])
            except (KeyError, ValueError):
                logger.warning(
                    "status_change %s has no valid event_time", status_change
                )
                continue
            validated_status_changes.append(status_change)

        return validated_status_changes

    def _validate_status_changes(self, status_changes):
        """Some preliminary checks/addenda"""
        validated_status_changes = []

        for status_change in status_changes:
            status_change["provider_id"] = uuid.UUID(status_change["provider_id"])
            status_change["device_id"] = device_id = uuid.UUID(
                status_change["device_id"]
            )

            # The list of event types and even the naming don't match between
            # the provider and agency APIs, so translate one to the other
            try:
                event_type_reason = status_change["event_type_reason"]
            except KeyError:  # Spec violation!
                logger.warning("Device %s has no event_type_reason", device_id)
                # Ignore just that status change to avoid rejecting the whole batch
                continue
            try:
                event = PROVIDER_REASON_TO_AGENCY_EVENT[event_type_reason]
                agency_event_type, agency_event_type_reason = (
                    event if len(event) == 2 else event + (None,)
                )
            except KeyError:  # Spec violation!
                logger.warning(
                    'Device %s has unknown "%s" event_type_reason',
                    device_id,
                    event_type_reason,
                )
                # Ignore just that status change to avoid rejecting the whole batch
                continue
            status_change["agency_event_type"] = agency_event_type
            status_change["agency_event_type_reason"] = agency_event_type_reason

            event_location = status_change["event_location"]
            if event_location:
                assert event_location["geometry"]["type"] == "Point"
                # Some providers may get the (lng, lat) order wrong
                if self.provider.api_configuration.get("swap_lng_lat"):
                    lat, lng = event_location["geometry"]["coordinates"][:2]
                    event_location["geometry"]["coordinates"][0] = lng
                    event_location["geometry"]["coordinates"][1] = lat
            else:  # Spec violation!
                logger.warning("Device %s has no event_location", device_id)
                # This time, accept a status change with no location

            validated_status_changes.append(status_change)

        return validated_status_changes

    def _create_missing_providers(self, status_changes):
        """Make sure all providers mentioned exist"""

        with_missing_providers = [
            status_change
            for status_change in status_changes
            if status_change["provider_id"] not in self.provider_uids
        ]

        if with_missing_providers:
            db_helpers.upsert_providers(
                (
                    _create_provider(status_change)
                    for status_change in with_missing_providers
                )
            )

            providers_added = [
                status_change["provider_id"] for status_change in with_missing_providers
            ]
            self.provider_uids.update(providers_added)

            logger.info(
                "Providers created: %s", ", ".join(str(uid) for uid in providers_added)
            )

    def _create_missing_devices(self, status_changes):
        """Make sure all devices mentioned exist"""

        with_missing_devices = [
            status_change
            for status_change in status_changes
            if status_change["device_id"] not in self.device_uids
        ]

        if with_missing_devices:
            db_helpers.upsert_devices(
                (
                    _create_device(status_change)
                    for status_change in with_missing_devices
                )
            )
            # Create fake register events to simulate device registration
            db_helpers.upsert_event_records(
                (
                    _create_register_event_record(status_change)
                    for status_change in with_missing_devices
                ),
                source=enums.EVENT_SOURCE.provider_api.name,
            )

            devices_added = [
                status_change["device_id"] for status_change in with_missing_devices
            ]
            self.device_uids.update(devices_added)

            logger.info(
                "Devices created: %s", ", ".join(str(uid) for uid in devices_added)
            )

    def _create_event_records(self, status_changes):
        """Now record the... records"""
        db_helpers.upsert_event_records(
            (_create_event_record(status_change) for status_change in status_changes),
            enums.EVENT_SOURCE.provider_api.name,
            # Timestamps are unique per device, ignore duplicates
            # Events already pushed by the provider will always have precedence
            on_conflict_update=False,
        )


def _create_provider(status_change):
    provider_id = status_change["provider_id"]
    try:
        name = status_change["provider_name"]
    except KeyError:  # Spec violation!
        logger.warning("Provider %s has no name", provider_id)
        name = ""

    return models.Provider(id=provider_id, name=name)


def _create_device(status_change):
    device_id = status_change["device_id"]
    identification_number = status_change["vehicle_id"]
    if not identification_number:  # Spec violation!
        logger.warning("Device %s has no identification number", device_id)
        identification_number = "test-%s" % str(device_id).split("-", 1)

    return models.Device(
        id=device_id,
        # Don't assume the device received belongs to the provider requested
        # The LA sandbox contains data for several providers
        provider_id=status_change["provider_id"],
        identification_number=identification_number,
        category=status_change["vehicle_type"],
        propulsion=status_change["propulsion_type"],
    )


def _create_event_record(status_change):
    properties = {"trip_id": status_change.get("associated_trip")}

    event_location = status_change["event_location"]
    if event_location:
        # GeoJSON Point Feature
        try:
            longitude, latitude, altitude = event_location["geometry"]["coordinates"]
        except ValueError:
            longitude, latitude = event_location["geometry"]["coordinates"]
            altitude = None
        point = geos.Point(longitude, latitude, altitude, srid=4326)
        properties["telemetry"] = {
            "timestamp": event_location["properties"]["timestamp"],
            "gps": {"lng": longitude, "lat": latitude},
            # No coordinates, no battery charge saved
            "battery_pct": status_change.get("battery_pct"),
        }
        if altitude:
            properties["telemetry"]["gps"]["altitude"] = altitude
    else:  # Spec violation!
        point = None

    first_saved_at = None
    # "Aggregation" layers store a recorded field, and we want to keep the same value
    if "recorded" in status_change and status_change["recorded"]:
        first_saved_at = utils.from_mds_timestamp(status_change["recorded"])

    return models.EventRecord(
        device_id=status_change["device_id"],
        timestamp=utils.from_mds_timestamp(status_change["event_time"]),
        point=point,
        event_type=status_change["agency_event_type"],
        event_type_reason=status_change["agency_event_type_reason"],
        properties=properties,
        first_saved_at=first_saved_at,
    )


def _create_register_event_record(status_change):
    """
    As the goal of the poller is to catch up with the history of a provider,
    simulate the registration of a device with a fake register event.

    This fake event is flagged in its properties to tell it apart from real ones.

    Should the device be unregistered and registered again according to the specs,
    don't delete the fake events in the past.
    """
    return models.EventRecord(
        device_id=status_change["device_id"],
        # Another event for the same device with the same timestamp will be rejected
        timestamp=utils.from_mds_timestamp(status_change["event_time"])
        - datetime.timedelta(milliseconds=1),
        event_type=enums.EVENT_TYPE.register.name,
        properties={"created_on_register": True},
        source=enums.EVENT_SOURCE.provider_api.name,
    )



# Generated by Django 2.1.11 on 2019-08-22 12:39

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [("mds", "0038_eventrecord_event_type_reason")]

    operations = [
        migrations.AlterField(
            model_name="eventrecord",
            name="event_type",
            field=mds.models.UnboundedCharField(
                choices=[
                    ("register", "Register"),
                    ("service_start", "Service start"),
                    ("service_end", "Service end"),
                    ("provider_drop_off", "Provider drop off"),
                    ("provider_pick_up", "Provider pick up"),
                    ("city_pick_up", "City pick up"),
                    ("reserve", "Reserve"),
                    ("cancel_reservation", "Cancel reservation"),
                    ("trip_start", "Trip start"),
                    ("trip_enter", "Trip enter"),
                    ("trip_leave", "Trip leave"),
                    ("trip_end", "Trip end"),
                    ("deregister", "Deregister"),
                    ("telemetry", "Received telemetry"),
                    ("battery_charged", "Battery charged"),
                    ("rebalance_drop_off", "Rebalance drop off"),
                    ("maintenance_drop_off", "Maintenance drop off"),
                    ("low_battery", "Low battery"),
                    ("maintenance", "Maintenance"),
                    ("rebalance_pick_up", "Rebalance pick up"),
                    ("maintenance_pick_up", "Maintenance pick up"),
                ]
            ),
        )
    ]



from rest_framework import serializers
from rest_framework import viewsets

from django.db.models import Q
from django.db.models.functions import Now

from mds import models
from mds import utils
from mds.apis import utils as apis_utils


class PolicySerializer(serializers.Serializer):
    name = serializers.CharField(help_text="Name of policy")
    policy_id = serializers.UUIDField(source="id", help_text="Unique ID of policy")
    provider_ids = serializers.SerializerMethodField(
        help_text=(
            "Providers for whom this policy is applicable "
            "(null or absent implies all Providers)"
        )
    )
    description = serializers.CharField(help_text="Description of policy")
    start_date = apis_utils.UnixTimestampMilliseconds(
        help_text="Beginning date/time of policy enforcement"
    )
    end_date = apis_utils.UnixTimestampMilliseconds(
        help_text="End date/time of policy enforcement"
    )
    published_date = apis_utils.UnixTimestampMilliseconds(
        help_text="Timestamp that the policy was published"
    )
    prev_policies = serializers.SerializerMethodField(
        help_text="Unique IDs of prior policies replaced by this one"
    )
    fixed_price = serializers.IntegerField()
    rules = serializers.JSONField(help_text="List of applicable rule elements")

    def get_provider_ids(self, policy):
        # Why we need a prefetch_related()!
        return [str(provider.id) for provider in policy.providers.all()]

    def get_prev_policies(self, policy):
        return [str(policy.id) for policy in policy.prev_policies.all()]


class PolicyViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = (
        models.Policy.objects.prefetch_related("providers", "prev_policies")
        .filter(published_date__isnull=False)  # Not drafts
        .order_by("start_date")
    )
    permission_classes = ()  # Public endpoint but results are restricted
    lookup_field = "id"
    serializer_class = PolicySerializer
    # TODO filter_backends

    def get_queryset(self):
        queryset = super().get_queryset()

        provider_id = getattr(self.request.user, "provider_id", None)
        if provider_id:
            # Filter for general-purpose policies,
            # or the ones written for this provider
            queryset = queryset.filter(
                Q(providers__isnull=True) | Q(providers=provider_id)
            )
            # The raw results will expose what other providers this policy applies to
            # Is this information leakage?
        else:
            # Only general-purpose policies to other users
            queryset = queryset.filter(providers__isnull=True)

        # Filter by date range
        start_time = self.request.GET.get("start_time")
        if start_time:
            start_time = utils.from_mds_timestamp(int(start_time))
        else:
            start_time = Now()
        range_q = Q(end_date__gte=start_time)

        end_time = self.request.GET.get("end_time")
        if end_time:
            end_time = utils.from_mds_timestamp(int(end_time))
            range_q &= Q(start_date__lte=end_time)

        queryset = queryset.filter(range_q)

        return queryset



# Generated by Django 2.1.4 on 2019-01-17 16:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("mds", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="area",
            name="color",
            field=models.CharField(
                default="#FFFFFF", help_text="hexa representation", max_length=8
            ),
        )
    ]



# Generated by Django 2.1.5 on 2019-01-28 15:25

import django.contrib.postgres.fields.hstore
from django.contrib.postgres.operations import HStoreExtension
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [("mds", "0006_provider_device_category")]

    operations = [
        HStoreExtension(),
        migrations.AddField(
            model_name="provider",
            name="authentication",
            field=django.contrib.postgres.fields.hstore.HStoreField(
                default=dict, verbose_name="API Authentication"
            ),
        ),
        migrations.AddField(
            model_name="provider",
            name="base_api_url",
            field=mds.models.UnboundedCharField(
                default="", verbose_name="Base API URL"
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-12 10:02

from django.db import migrations, models


def fill_default_aggregator_for(apps, schema_editor):
    Application = apps.get_model("authent", "Application")
    Provider = apps.get_model("mds", "Provider")

    for application in Application.objects.all():
        if application.owner:
            try:
                provider = Provider.objects.get(pk=application.owner)
                application.aggregator_for.add(provider)
            except Provider.DoesNotExist:
                pass


class Migration(migrations.Migration):
    atomic = False

    dependencies = [("authent", "0003_pre_add_aggregator_for_application")]

    operations = [
        migrations.RunPython(fill_default_aggregator_for, migrations.RunPython.noop)
    ]



# Generated by Django 2.2 on 2019-07-03 09:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("mds", "0031_post_eventrecord_saved_at_index")]

    operations = [
        migrations.AddField(
            model_name="device",
            name="saved_at",
            field=models.DateTimeField(db_index=True, null=True),
        )
    ]



from uuid import UUID

from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from . import models


def is_uuid(uuid_string, version=4):
    try:
        UUID(uuid_string, version=version)
    except ValueError:
        return False
    return True


@admin.register(models.Provider)
class ProviderAdmin(admin.ModelAdmin):
    list_display = ["id", "name", "operator"]
    list_filter = ["operator"]
    ordering = ["name"]


@admin.register(models.Device)
class DeviceAdmin(admin.ModelAdmin):
    list_display = ["id", "provider_name", "identification_number", "category"]
    list_filter = ["provider", "category"]
    search_fields = ["id", "identification_number"]
    list_select_related = ["provider"]

    def provider_name(self, obj):
        return obj.provider.name

    provider_name.short_description = "Provider"


@admin.register(models.EventRecord)
class EventRecordAdmin(admin.ModelAdmin):
    list_display = [
        "saved_at",
        "timestamp",
        "provider",
        "device_id",
        "event_type",
        "event_type_reason",
    ]
    list_filter = ["device__provider", "event_type", "event_type_reason"]
    list_select_related = ["device__provider"]
    search_fields = ["device__id", "device__identification_number"]

    def get_search_results(self, request, queryset, search_term):
        if not search_term:
            return super().get_search_results(request, queryset, search_term)
        custom_queryset = get_devices_queryset_search_results(self, search_term)
        return super().get_search_results(request, custom_queryset, search_term)

    def provider(self, obj):
        return obj.device.provider.name

    def device_id(self, obj):
        return obj.device.id


# to use when searching for devices in get_search_results
# as a relationship to self.model
def get_devices_queryset_search_results(self, search_term):
    custom_queryset = self.model.objects.select_related(*self.list_select_related)
    if is_uuid(search_term):
        return custom_queryset.filter(device_id=search_term)
    else:
        return custom_queryset.filter(device__identification_number=search_term)


@admin.register(models.Area)
class AreaAdmin(admin.ModelAdmin):
    list_display = ["id", "label"]
    ordering = ["label"]


@admin.register(models.Polygon)
class PolygonAdmin(admin.ModelAdmin):
    list_display = ["id", "label"]
    ordering = ["label"]


class PublishedFilter(admin.SimpleListFilter):
    title = _("published")
    parameter_name = "is_published"

    def lookups(self, request, model_admin):
        return [("0", _("No")), ("1", _("Yes"))]

    def queryset(self, request, queryset):
        if self.value() == "0":
            queryset = queryset.filter(published_date__isnull=True)
        elif self.value() == "1":
            queryset = queryset.filter(published_date__isnull=False)
        return queryset


@admin.register(models.Policy)
class PolicyAdmin(admin.ModelAdmin):
    list_display = ["name", "start_date", "end_date", "is_published"]
    list_filter = ["start_date", "end_date", PublishedFilter, "providers"]
    search_fields = ["id", "name", "description"]
    ordering = ["start_date"]

    def is_published(self, policy):
        return bool(policy.published_date)

    is_published.boolean = True


@admin.register(models.Compliance)
class ComplianceAdmin(admin.ModelAdmin):
    list_display = ["id", "policy_name", "start_date", "end_date"]
    list_filter = ["start_date", "end_date"]
    search_fields = ["id", "policy_id", "vehicle_id"]
    ordering = ["start_date"]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("policy")

    def policy_name(self, compliance):
        return compliance.policy.name



# Generated by Django 2.1.7 on 2019-03-28 16:19

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0016_agency_api_configuration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='device',
            name='category',
            field=mds.models.UnboundedCharField(choices=[('bicycle', 'Bicycle'), ('scooter', 'Scooter'), ('car', 'Car')]),
        ),
        migrations.AlterField(
            model_name='provider',
            name='device_category',
            field=mds.models.UnboundedCharField(choices=[('bicycle', 'Bicycle'), ('scooter', 'Scooter'), ('car', 'Car')]),
        ),
    ]



# Generated by Django 2.1.10 on 2019-07-23 17:26

from django.db import migrations


def fill_provider_configuration(apps, schema_editor):
    Provider = apps.get_model("mds", "Provider")

    for provider in Provider.objects.all():
        prev_api_config = provider.api_configuration
        prev_start_time_field = prev_api_config.get("start_time_field", "start_time")

        # First, we update the api_configuration
        provider.api_configuration["polling_cursor"] = prev_start_time_field
        if provider.api_configuration.get("start_time_field"):
            del provider.api_configuration["start_time_field"]

        # Next, we update the timestamp of the cursor fields
        if prev_start_time_field == "start_time":
            provider.last_event_time_polled = provider.last_start_time_polled
        elif prev_start_time_field == "start_recorded":
            provider.last_recorded_polled = provider.last_start_time_polled

        provider.save(
            update_fields=[
                "api_configuration",
                "last_event_time_polled",
                "last_recorded_polled",
            ]
        )


class Migration(migrations.Migration):
    atomic = False

    dependencies = [("mds", "0035_pre_add_fields")]

    operations = [
        migrations.RunPython(fill_provider_configuration, migrations.RunPython.noop)
    ]




import datetime

from django.conf import settings


AUTHENT_RSA_PRIVATE_KEY = getattr(settings, "AUTHENT_RSA_PRIVATE_KEY", "")
AUTHENT_SECRET_KEY = getattr(settings, "AUTHENT_SECRET_KEY", "")
assert not all([AUTHENT_SECRET_KEY, AUTHENT_RSA_PRIVATE_KEY]), (
    "You must define either settings.AUTHENT_RSA_PRIVATE_KEY "
    "or settings.AUTHENT_SECRET_KEY and not both."
)
AUTHENT_LONG_LIVED_TOKEN_DURATION = getattr(
    settings, "AUTHENT_LONG_LIVED_TOKEN_DURATION", datetime.timedelta(days=365)
)



# Generated by Django 2.1.11 on 2019-08-16 14:27

import datetime

from django.db import migrations, transaction
from django.db.models import Q

from mds.provider_mapping import (
    OLD_PROVIDER_REASON_TO_AGENCY_EVENT,
    OLD_TO_NEW_AGENCY_EVENT,
    PROVIDER_REASON_TO_AGENCY_EVENT,
    get_new_event_from_old,
    get_old_event_from_new,
    get_same_mapping_event,
)


def full_batch_apply(
    function, qs, collation=lambda a, b: a and b, initial_value=None, **kwargs
):
    batch_size = 1000
    try:
        first_pk = qs.order_by("pk").values_list("pk", flat=True)[0]
    except IndexError:
        return initial_value
    last_pk = qs.order_by("-pk").values_list("pk", flat=True)[0]
    start = max(1, first_pk)

    for i in range(start, last_pk + 1, batch_size):
        queryset = qs.filter(pk__gte=i, pk__lt=min(last_pk + 1, i + batch_size))
        try:
            with transaction.atomic():
                initial_value = collation(initial_value, function(queryset, **kwargs))
        except Exception:
            raise

    return initial_value


def fill_event_type_and_reason(apps, schema_editor):
    EventRecord = apps.get_model("mds", "EventRecord")

    for agency_event in OLD_PROVIDER_REASON_TO_AGENCY_EVENT.values():
        new_agency_event = get_new_event_from_old(agency_event)
        if new_agency_event in get_same_mapping_event():
            # We do not want to apply the migration for the events
            # that are in both mappings
            continue

        event_type, event_type_reason = (
            new_agency_event
            if len(new_agency_event) == 2
            else new_agency_event + (None,)
        )

        qs = EventRecord.objects.filter(
            Q(event_type=agency_event) & Q(event_type_reason__isnull=True)
        )

        def fill(qs):
            if event_type_reason:
                qs.update(event_type=event_type, event_type_reason=event_type_reason)
            else:
                qs.update(event_type=event_type)

        full_batch_apply(fill, qs)


def reverse_fill_event_type_and_reason(apps, schema_editor):
    EventRecord = apps.get_model("mds", "EventRecord")

    for agency_event in PROVIDER_REASON_TO_AGENCY_EVENT.values():
        if agency_event in get_same_mapping_event():
            # We do not want to apply the reverse migration to the events
            # that are in both mappings
            continue

        old_event = get_old_event_from_new(agency_event)[0]

        event_type, event_type_reason = (
            agency_event if len(agency_event) == 2 else agency_event + (None,)
        )

        if event_type_reason:
            qs = EventRecord.objects.filter(
                Q(event_type=event_type) & Q(event_type_reason=event_type_reason)
            )
        else:
            qs = EventRecord.objects.filter(
                Q(event_type=event_type) & Q(event_type_reason__isnull=True)
            )

        def fill(qs):
            qs.update(event_type=old_event, event_type_reason=None)

        full_batch_apply(fill, qs)


class Migration(migrations.Migration):
    atomic = False

    dependencies = [("mds", "0040_change_event_record_event_type_options")]

    operations = [
        migrations.RunPython(
            fill_event_type_and_reason, reverse_fill_event_type_and_reason
        )
    ]



# Generated by Django 2.2 on 2019-06-19 14:38

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("mds", "0028_pre_provider_operator")]

    operations = [
        migrations.RunSQL(
            """
            UPDATE mds_provider SET operator = (
                CASE WHEN name IN ('LADOT', 'LADOT Sandbox') THEN false
                ELSE true
                END
            )
            """,
            migrations.RunSQL.noop,
        )
    ]



# Generated by Django 2.2.5 on 2019-09-12 10:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mds", "0044_add_compliance_list"),
        ("authent", "0002_auto_20190124_0928"),
    ]

    operations = [
        migrations.AddField(
            model_name="application",
            name="aggregator_for",
            field=models.ManyToManyField(
                help_text="Provider the application is allowed to write for",
                to="mds.Provider",
            ),
        ),
        # Lingering migration, not related to this PR
        migrations.AlterField(
            model_name="application",
            name="owner",
            field=models.UUIDField(
                blank=True,
                db_index=True,
                help_text="Unique identifier for the owner of the application",
                null=True,
            ),
        ),
    ]



"""A few helpers around DRF."""
import datetime
import json

from django.contrib.gis import geos
from django_filters import rest_framework as filters
from rest_framework import pagination
from rest_framework import serializers
from rest_framework.response import Response

from rest_framework import status
from rest_framework.exceptions import APIException
from django.utils.translation import pgettext_lazy

import mds.utils


# Errors #######################################################


class AlreadyRegisteredError(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_detail = pgettext_lazy("API Error", "Already registered")
    default_code = "already_registered"


# Pagination ###################################################


class LimitOffsetPagination(pagination.LimitOffsetPagination):
    default_limit = 100


# Viewsets #####################################################


class MultiSerializerViewSetMixin:
    serializers_mapping = {}

    def _create(self, request, *args, **kwargs):
        """Patch of rest_framework.mixins.CreateModelMixin.create.

        This patch works with our Request/Response serializers.
        """
        request_serializer = self.get_serializer(
            data=request.data, context={"request_or_response": "request"}
        )
        request_serializer.is_valid(raise_exception=True)
        instance = request_serializer.save()
        response_serializer = self.get_serializer(
            instance=instance, context={"request_or_response": "response"}
        )
        return Response(response_serializer.data, status=201)

    def _update(self, request, *args, **kwargs):
        """Patch of rest_framework.mixins.UpdateModelMixin.create.

        This patch works with our Request/Response serializers.
        """
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        request_serializer = self.get_serializer(
            instance,
            data=request.data,
            partial=partial,
            context={"request_or_response": "request"},
        )
        request_serializer.is_valid(raise_exception=True)
        request_serializer.save()

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        response_serializer = self.get_serializer(
            instance=instance, context={"request_or_response": "response"}
        )
        return Response(response_serializer.data)

    def get_serializer(self, *args, **kwargs):
        context = kwargs.get("context", {})
        context.update(self.get_serializer_context())
        kwargs["context"] = context
        request_or_response = context.get("request_or_response", "response")
        action = self.action
        if action == "partial_update":
            action = "update"
        serializer = self.serializers_mapping.get(action, {}).get(request_or_response)
        if serializer:
            return serializer(*args, **kwargs)
        return super().get_serializer(*args, **kwargs)


# Serializers ##################################################


class EmptyResponseSerializer(serializers.Serializer):
    pass


class BaseGeometrySerializer(serializers.Serializer):
    def to_representation(self, value):
        if isinstance(value, dict) or value is None:
            return value
        return json.loads(value.geojson)

    def to_internal_value(self, value):
        if not isinstance(value, dict):
            value = json.loads(value)
        return value


class PointSerializer(BaseGeometrySerializer):
    type = serializers.ChoiceField(["Point"])
    coordinates = serializers.ListField(
        child=serializers.FloatField(), min_length=2, max_length=3
    )  # could include altitude


class MultiPointSerializer(BaseGeometrySerializer):
    type = serializers.ChoiceField(["MultiPoint"])
    coordinates = serializers.ListField(
        child=serializers.ListField(child=serializers.FloatField())
    )


class PolygonSerializer(BaseGeometrySerializer):
    type = serializers.ChoiceField(["Polygon", "MultiPolygon"])
    coordinates = serializers.ListField(
        child=serializers.ListField(
            child=serializers.ListField(child=serializers.FloatField())
        )
    )


class UnixTimestampMilliseconds(serializers.IntegerField):
    def to_representation(self, value: datetime.datetime):
        return mds.utils.to_mds_timestamp(value)

    def to_internal_value(self, value: int):
        return mds.utils.from_mds_timestamp(value)


# Schema #########################################################
# Filters #########################################################


class UUIDInFilter(filters.BaseInFilter, filters.UUIDFilter):
    pass


class ChoicesInFilter(filters.BaseInFilter, filters.ChoiceFilter):
    pass


class DateTimeRangeOverlapFilter(filters.IsoDateTimeFromToRangeFilter):
    # Handles datetime range overlap filter
    def filter(self, qs, value):
        if not value:
            return qs

        lookup = "%s__%s" % (self.field_name, "overlap")
        return self.get_method(qs)(**{lookup: (value.start, value.stop)})


# Others #########################################################


def get_geos_multipolygon(geom_str_repr):
    """
      Retrieve a GEOS MultiPolygon from string
    """
    geos_obj = geos.GEOSGeometry(geom_str_repr)
    if geos_obj and not isinstance(geos_obj, geos.MultiPolygon):
        return geos.MultiPolygon(geos_obj)
    return geos_obj



# Generated by Django 2.2 on 2019-05-17 08:45

import django.contrib.postgres.fields.jsonb
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [("mds", "0022_provider_colors")]

    operations = [
        migrations.AlterField(
            model_name="provider",
            name="agency_api_authentication",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True,
                default=mds.models.agency_api_authentication_default,
                verbose_name="Agency API Authentication",
            ),
        ),
        migrations.AlterField(
            model_name="provider",
            name="agency_api_configuration",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True,
                default=mds.models.agency_api_configuration_default,
                verbose_name="Agency API Configuration",
            ),
        ),
        migrations.AlterField(
            model_name="provider",
            name="api_authentication",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True,
                default=mds.models.provider_api_authentication_default,
                verbose_name="Provider API Authentication",
            ),
        ),
        migrations.AlterField(
            model_name="provider",
            name="api_configuration",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True,
                default=mds.models.provider_api_configuration_default,
                verbose_name="Provider API Configuration",
            ),
        ),
    ]



# Generated by Django 2.1.11 on 2019-09-09 14:34

import django.contrib.postgres.fields.jsonb
import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0043_policy'),
    ]

    operations = [
        migrations.CreateModel(
            name='Compliance',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('rule', models.UUIDField(default=uuid.uuid4)),
                ('geography', models.UUIDField(default=uuid.uuid4)),
                ('start_date', models.DateTimeField()),
                ('end_date', models.DateTimeField(blank=True, null=True)),
            ],
        ),
        migrations.AlterField(
            model_name='policy',
            name='prev_policies',
            field=models.ManyToManyField(blank=True, related_name='_policy_prev_policies_+', to='mds.Policy'),
        ),
        migrations.AlterField(
            model_name='policy',
            name='providers',
            field=models.ManyToManyField(blank=True, to='mds.Provider'),
        ),
        migrations.AlterField(
            model_name='policy',
            name='rules',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=list, encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
        migrations.AddField(
            model_name='compliance',
            name='policy',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='compliances', to='mds.Policy'),
        ),
        migrations.AddField(
            model_name='compliance',
            name='vehicle',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='compliances', to='mds.Device'),
        ),
    ]



# Generated by Django 2.1.12 on 2019-09-19 13:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0044_add_compliance_list'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='policy',
            options={'verbose_name_plural': 'policies'},
        ),
        migrations.AddField(
            model_name='policy',
            name='fixed_price',
            field=models.IntegerField(default=0),
        ),
    ]



# Generated by Django 2.2 on 2019-08-20 14:10

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import mds.models
import uuid


class Migration(migrations.Migration):
    deploy_phase = "pre_deploy"

    dependencies = [("mds", "0042_revert_0039_event_type")]

    operations = [
        migrations.CreateModel(
            name="Policy",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("name", mds.models.UnboundedCharField()),
                ("description", mds.models.UnboundedCharField(blank=True, default="")),
                ("start_date", models.DateTimeField()),
                ("end_date", models.DateTimeField(blank=True, null=True)),
                ("published_date", models.DateTimeField(blank=True, null=True)),
                ("rules", django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                (
                    "prev_policies",
                    models.ManyToManyField(
                        related_name="_policy_prev_policies_+", to="mds.Policy"
                    ),
                ),
                ("providers", models.ManyToManyField(to="mds.Provider")),
            ],
        )
    ]



import json
import logging
import threading
import urllib.parse

from django.core.cache import caches

from cryptography import fernet
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from retrying import retry

from .settings import POLLER_TOKEN_CACHE, POLLER_TOKEN_ENCRYPTION_KEY


CACHE_KEY_PATTERN = "mds:oauth2-token:%s"  # Provider ID added

cache = caches[POLLER_TOKEN_CACHE]

logger = logging.getLogger(__name__)


def _encrypt_token(token):
    encoded = json.dumps(token).encode("utf8")  # Fernet needs bytes
    return fernet.Fernet(POLLER_TOKEN_ENCRYPTION_KEY).encrypt(encoded)


def _decrypt_token(encrypted):
    try:
        encoded = fernet.Fernet(POLLER_TOKEN_ENCRYPTION_KEY).decrypt(encrypted)
    except (fernet.InvalidToken, TypeError):  # The encryption key, not our token!
        logger.debug("Invalid token stored in the cache", exc_info=True)
        return None
    else:
        return json.loads(encoded)


class OAuth2Store(threading.local):
    def __init__(self, provider):
        self.provider = provider

    def get_client(self):
        """Authenticate using the Backend Application Flow from OAuth2."""
        client_id = self.provider.api_authentication["client_id"]
        token = self._get_access_token()
        client = OAuth2Session(client_id, token=token)
        return client

    def _get_access_token(self):
        """Fetch and cache the token."""
        key = CACHE_KEY_PATTERN % self.provider.pk
        token = _decrypt_token(cache.get(key))
        if token:
            logger.debug("Token already in cache")
            return token
        token = self._fetch_access_token()
        cache.set(key, _encrypt_token(token), timeout=token["expires_in"] - 10)
        logger.debug("New token stored in cache")
        return token

    @retry(stop_max_attempt_number=2)
    def _fetch_access_token(self):
        """Fetch a new access token from the provider"""
        client_id = self.provider.api_authentication["client_id"]
        client_secret = self.provider.api_authentication["client_secret"]
        client = BackendApplicationClient(client_id=client_id)
        session = OAuth2Session(client=client)

        # The URL to fetch the token maybe on another base URL
        token_url = self.provider.oauth2_url
        if not token_url:
            token_url = urllib.parse.urljoin(
                self.provider.base_api_url, "/oauth2/token"
            )
            if self.provider.api_configuration.get("trailing_slash"):
                token_url += "/"

        # Provider-specific parameters to request the token
        kwargs = {}
        try:
            kwargs.update(self.provider.api_configuration["token_params"])
        except KeyError:
            pass

        token = session.fetch_token(
            token_url=token_url,
            client_id=client_id,
            client_secret=client_secret,
            **kwargs
        )
        return token

    def flush_token(self):
        key = CACHE_KEY_PATTERN % self.provider.pk
        logger.debug("Token deleted from cache")
        cache.delete(key)



# Generated by Django 2.2 on 2019-05-23 13:18

import django.contrib.postgres.fields.jsonb
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0024_provider_device_category_deprecation'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='provider',
            name='device_category',
        ),
        migrations.AlterField(
            model_name='provider',
            name='colors',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, validators=[mds.models.colors_validator], verbose_name='colors for distinguishing provider icons'),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-25 13:49

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0017_change_bike_to_bicycle_in_choices_field'),
    ]

    operations = [
        migrations.AlterField(
            model_name='provider',
            name='base_api_url',
            field=mds.models.UnboundedCharField(blank=True, default='', verbose_name='Base Provider API URL'),
        ),
        migrations.AlterField(
            model_name='provider',
            name='oauth2_url',
            field=mds.models.UnboundedCharField(blank=True, default='', verbose_name='OAuth2 URL (if different)'),
        ),
    ]



# Generated by Django 2.1.7 on 2019-05-17 13:48

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("mds", "0021_rename_event_source")]

    operations = [
        migrations.AddField(
            model_name="provider",
            name="colors",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True,
                default=dict,
                verbose_name="colors for distinguishing provider icons",
            ),
        )
    ]



import logging

import datetime
from typing import List

from django.apps import apps
from django.core.cache import cache
from django.utils import timezone


from mds.models import Provider
from mds.authent import models, generators

logger = logging.getLogger(__name__)


class MDSAuthentException(Exception):
    pass


class NoApplicationForOwner(MDSAuthentException):
    pass


class UnknownToken(MDSAuthentException):
    pass


def get_long_lived_token(owner, duration):
    application = models.Application.objects.filter(owner=owner).last()
    if not application:
        raise NoApplicationForOwner()

    token_duration = datetime.timedelta(seconds=duration)
    token = generators.generate_jwt(application, token_duration, save=True)
    return token


def revoke_long_lived_token(token):
    token_qs = models.AccessToken.objects.filter(token=token)
    stored_token = token_qs.last()
    if not stored_token:
        raise UnknownToken()

    if not stored_token.revoked_after:
        _revoke_tokens(token_qs)
    return stored_token.revoked_after


def get_revocation_list() -> List[str]:
    """Get the list of revoked tokens.

    The code below could be replaced by an API call if the authentication
    server and the resource server are split.
    see (https://tools.ietf.org/id/draft-gpujol-oauth-atrl-00.html)
    """
    cache_key = "authent:revocation_list"
    cached = cache.get(cache_key)
    if cached:
        return cached

    try:
        apps.get_app_config("authent")
    except LookupError:
        raise NotImplementedError(
            "Calling this function requires mds.authent in INSTALLED_APPS."
        )

    token_ids = [
        str(t)
        for t in models.AccessToken.objects.filter(
            revoked_after__lt=timezone.now()
        ).values_list("jti", flat=True)
    ]
    cache.set(cache_key, token_ids, timeout=60)  # seconds
    return token_ids


def create_application(name, owner=None, grant=None, scopes=None):
    grant = grant or models.Application.GRANT_CLIENT_CREDENTIALS
    app, created = models.Application.objects.get_or_create(
        name=name,
        client_type=models.Application.CLIENT_CONFIDENTIAL,
        authorization_grant_type=grant,
        scopes=scopes,
        owner=owner,
    )
    try:
        provider = Provider.objects.get(pk=owner)
        app.aggregator_for.add(provider)
    except Provider.DoesNotExist:
        # We raise a warning but continue with the creation of the client credentials
        logger.warning(f"Could not set app permissions for owner {owner}")
        pass
    return {
        "name": name,
        "client_id": app.client_id,
        "client_secret": app.client_secret,
        "client_type": models.Application.CLIENT_CONFIDENTIAL,
        "grant_type": grant,
        "scopes": scopes,
        "created": created,
    }


def revoke_application(owner_id):
    app_qs = models.Application.objects.filter(owner=owner_id)
    updated = app_qs.update(scopes=[])
    if not updated:
        raise NoApplicationForOwner()
    _revoke_tokens(
        models.AccessToken.objects.filter(
            application_id__in=app_qs.values_list("id", flat=True)
        )
    )


def delete_application(owner_id):
    deleted, _ = models.Application.objects.filter(owner=owner_id).delete()
    if not deleted:
        raise NoApplicationForOwner()


def _revoke_tokens(token_qs):
    token_qs.update(revoked_after=timezone.now())



# Generated by Django 2.1.5 on 2019-01-29 16:41

import django.contrib.postgres.fields.hstore
import django.contrib.postgres.fields.jsonb
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [("mds", "0007_provider_api")]

    operations = [
        migrations.RenameField(
            model_name="provider",
            old_name="authentication",
            new_name="api_authentication",
        ),
        migrations.AlterField(
            model_name="provider",
            name="api_authentication",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=mds.models.provider_api_authentication_default,
                verbose_name="API Authentication",
            ),
        ),
        migrations.AddField(
            model_name="provider",
            name="api_configuration",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=mds.models.provider_api_configuration_default,
                verbose_name="API Configuration",
            ),
        ),
        migrations.AddField(
            model_name="provider",
            name="oauth2_url",
            field=mds.models.UnboundedCharField(
                default="", verbose_name="OAuth2 URL (if different)"
            ),
        ),
    ]



import json
import types

from django.db import connection
from rest_framework.utils import encoders


def upsert_providers(providers: types.GeneratorType):
    """
    Using "upsert" to create providers.

    Conflicts are always ignored.
    """

    def serialize(provider):
        provider.clean()
        return {
            "id": provider.id,
            "name": provider.name,
            "base_api_url": provider.base_api_url,
            "oauth2_url": provider.oauth2_url,
            # JSONField doesn't use Django JSON encoder by default
            "api_authentication": json.dumps(provider.api_authentication),
            "api_configuration": json.dumps(provider.api_configuration),
            "agency_api_authentication": json.dumps(provider.agency_api_authentication),
            "agency_api_configuration": json.dumps(provider.agency_api_configuration),
            "colors": json.dumps(provider.colors),
            "operator": provider.operator,
        }

    query = """
        INSERT INTO mds_provider (
            id,
            name,
            base_api_url,
            oauth2_url,
            api_authentication,
            api_configuration,
            agency_api_authentication,
            agency_api_configuration,
            colors,
            operator
        ) VALUES (
            %(id)s,
            %(name)s,
            %(base_api_url)s,
            %(oauth2_url)s,
            %(api_authentication)s,
            %(api_configuration)s,
            %(agency_api_authentication)s,
            %(agency_api_configuration)s,
            %(colors)s,
            %(operator)s
        ) ON CONFLICT DO NOTHING
    """

    with connection.cursor() as cursor:
        cursor.executemany(query, (serialize(provider) for provider in providers))


def upsert_devices(devices: types.GeneratorType):
    """
    Using "upsert" to create devices.

    Conflicts are always ignored.
    """

    def serialize(device):
        device.clean()
        return {
            "id": device.id,
            "provider_id": device.provider_id,
            "registration_date": device.registration_date,
            "identification_number": device.identification_number,
            "category": device.category,
            "model": device.model,
            "propulsion": device.propulsion,
            "manufacturer": device.manufacturer,
            "dn_status": device.dn_status,
        }

    query = """
        INSERT INTO mds_device (
            id,
            provider_id,
            registration_date,
            identification_number,
            category,
            model,
            propulsion,
            manufacturer,
            dn_status,
            saved_at
        ) VALUES (
            %(id)s,
            %(provider_id)s,
            %(registration_date)s,
            %(identification_number)s,
            %(category)s,
            %(model)s,
            %(propulsion)s,
            %(manufacturer)s,
            %(dn_status)s,
            CURRENT_TIMESTAMP
        ) ON CONFLICT DO NOTHING
    """

    with connection.cursor() as cursor:
        cursor.executemany(query, (serialize(device) for device in devices))


def upsert_event_records(
    event_records: types.GeneratorType, source: str, on_conflict_update=False
):
    """
    Using "upsert" to create event or telemetry records.

    Records pushed by providers have precedence over the ones we pull from them.
    So we overwrite any existing record with the same timestamp for a given device.

    Lines will be created or updated with the save time of the transaction.

    Args:
        event_records: list of EventRecord
        source: enums.EVENT_SOURCE
        on_conflict_update: ignore duplicates (default) or overwrite
    """

    def serialize(event_record):
        event_record.clean()
        return {
            "timestamp": event_record.timestamp,
            "device_id": str(event_record.device_id),
            "point": event_record.point.ewkt if event_record.point else None,
            "event_type": event_record.event_type,
            "event_type_reason": event_record.event_type_reason,
            # The same encoder as in the model
            "properties": json.dumps(event_record.properties, cls=encoders.JSONEncoder),
            "source": source,
            "first_saved_at": event_record.first_saved_at,
        }

    query = """
        INSERT INTO mds_eventrecord (
            device_id,
            timestamp,
            point,
            event_type,
            event_type_reason,
            properties,
            source,
            first_saved_at,
            saved_at
        ) VALUES (
            %(device_id)s,
            %(timestamp)s,
            %(point)s,
            %(event_type)s,
            %(event_type_reason)s,
            %(properties)s,
            %(source)s,
            %(first_saved_at)s,
            current_timestamp
        )
        """
    if on_conflict_update:
        query += """
            ON CONFLICT (device_id, timestamp) DO UPDATE SET
                point = EXCLUDED.point,
                event_type = EXCLUDED.event_type,
                event_type_reason = EXCLUDED.event_type_reason,
                properties = EXCLUDED.properties,
                source = EXCLUDED.source,
                first_saved_at = EXCLUDED.first_saved_at,
                saved_at = current_timestamp
            """
    else:
        query += """
            ON CONFLICT DO NOTHING
            """

    with connection.cursor() as cursor:
        cursor.executemany(query, (serialize(record) for record in event_records))



# Generated by Django 2.2 on 2019-06-19 14:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("mds", "0029_post_provider_operator_fill")]

    operations = [
        migrations.AlterField(
            model_name="provider",
            name="operator",
            field=models.BooleanField(default=True),
        )
    ]



# Generated by Django 2.1.3 on 2019-04-18 13:49

import django.contrib.gis.db.models.fields
from django.db import migrations, models
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0019_delete_creation_deletion_date_fields_in_area_and_polygon'),
    ]

    operations = [
        migrations.AlterField(
            model_name='area',
            name='color',
            field=mds.models.UnboundedCharField(default='#FFFFFF', help_text='hexa representation'),
        ),
        migrations.AlterField(
            model_name='device',
            name='dn_battery_pct',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='device',
            name='dn_gps_point',
            field=django.contrib.gis.db.models.fields.PointField(blank=True, null=True, srid=4326),
        ),
        migrations.AlterField(
            model_name='device',
            name='dn_gps_timestamp',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='device',
            name='dn_status',
            field=mds.models.UnboundedCharField(choices=[('available', 'Available'), ('reserved', 'Reserved'), ('unavailable', 'Unavailable'), ('removed', 'Removed'), ('trip', 'Trip'), ('elsewhere', 'Elsewhere'), ('inactive', 'Inactive'), ('unknown', 'Unknown')], default='unknown'),
        ),
        migrations.AlterField(
            model_name='device',
            name='manufacturer',
            field=mds.models.UnboundedCharField(blank=True, default=''),
        ),
        migrations.AlterField(
            model_name='device',
            name='model',
            field=mds.models.UnboundedCharField(blank=True, default=''),
        ),
        migrations.AlterField(
            model_name='device',
            name='year_manufactured',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='eventrecord',
            name='point',
            field=django.contrib.gis.db.models.fields.PointField(blank=True, null=True, srid=4326),
        ),
        migrations.AlterField(
            model_name='eventrecord',
            name='source',
            field=mds.models.UnboundedCharField(choices=[('push', 'push'), ('pull', 'pull')], default='push'),
        ),
        migrations.AlterField(
            model_name='provider',
            name='name',
            field=mds.models.UnboundedCharField(blank=True, default=''),
        ),
    ]



import enum

from django.utils.translation import pgettext_lazy


def choices(enum):
    """Convert an Enum to a Django-style "choices" iterator.
    """
    return [(member.name, member.value) for member in enum]


DEVICE_STATUS = enum.Enum(
    "Device status",
    [
        ("available", pgettext_lazy("Device status", "Available")),
        ("reserved", pgettext_lazy("Device status", "Reserved")),
        ("unavailable", pgettext_lazy("Device status", "Unavailable")),
        ("removed", pgettext_lazy("Device status", "Removed")),
        ("trip", pgettext_lazy("Device status", "Trip")),
        ("elsewhere", pgettext_lazy("Device status", "Elsewhere")),
        ("inactive", pgettext_lazy("Device status", "Inactive")),
        ("unknown", pgettext_lazy("Device status", "Unknown")),
    ],
)

DEVICE_CATEGORY = enum.Enum(
    "Device category",
    [
        ("bicycle", pgettext_lazy("Device category", "Bicycle")),
        ("scooter", pgettext_lazy("Device category", "Scooter")),
        ("car", pgettext_lazy("Device category", "Car")),
    ],
)
DEVICE_PROPULSION = enum.Enum(
    "Device propulsion",
    [
        ("human", pgettext_lazy("Device propulsion", "Human")),
        ("electric_assist", pgettext_lazy("Device propulsion", "Electric Assist")),
        ("electric", pgettext_lazy("Device propulsion", "Electric")),
        ("combustion", pgettext_lazy("Device propulsion", "Combustion")),
    ],
)

EVENT_TYPE = enum.Enum(
    "Event type",
    [
        # The first events are the one listed in the MDS agency API, in order:
        # https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/agency
        ("register", pgettext_lazy("Event type", "Register")),
        ("service_start", pgettext_lazy("Event type", "Service start")),
        ("service_end", pgettext_lazy("Event type", "Service end")),
        ("provider_drop_off", pgettext_lazy("Event type", "Provider drop off")),
        ("provider_pick_up", pgettext_lazy("Event type", "Provider pick up")),
        ("city_pick_up", pgettext_lazy("Event type", "City pick up")),
        ("reserve", pgettext_lazy("Event type", "Reserve")),
        ("cancel_reservation", pgettext_lazy("Event type", "Cancel reservation")),
        ("trip_start", pgettext_lazy("Event type", "Trip start")),
        ("trip_enter", pgettext_lazy("Event type", "Trip enter")),
        ("trip_leave", pgettext_lazy("Event type", "Trip leave")),
        ("trip_end", pgettext_lazy("Event type", "Trip end")),
        ("deregister", pgettext_lazy("Event type", "Deregister")),
        # this last event is in neither of the MDS spec
        ("telemetry", pgettext_lazy("Event type", "Received telemetry")),
        # This may be added in a revision of the agency API specs
        ("battery_charged", pgettext_lazy("Event type", "Battery charged")),
    ],
)


# To make sure we still support the old version of the Agency API
# when we receive events.
OLD_EVENT_TYPE = enum.Enum(
    "Event type",
    [
        ("rebalance_drop_off", pgettext_lazy("Event type", "Rebalance drop off")),
        ("maintenance_drop_off", pgettext_lazy("Event type", "Maintenance drop off")),
        ("low_battery", pgettext_lazy("Event type", "Low battery")),
        ("maintenance", pgettext_lazy("Event type", "Maintenance")),
        ("rebalance_pick_up", pgettext_lazy("Event type", "Rebalance pick up")),
        ("maintenance_pick_up", pgettext_lazy("Event type", "Maintenance pick up")),
    ],
)


# event_type_reason(s) listed in the provider API but not in the agency API
# https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/provider
PROVIDER_EVENT_TYPE_REASON = enum.Enum(
    "Provider event type reason",
    [
        # available #
        ("service_start", pgettext_lazy("Event type", "Service start")),
        ("user_drop_off", pgettext_lazy("Event type", "User drop off")),
        ("rebalance_drop_off", pgettext_lazy("Event type", "Rebalance drop off")),
        ("maintenance_drop_off", pgettext_lazy("Event type", "Maintenance drop off")),
        ("agency_drop_off", pgettext_lazy("Event type", "Agency drop off")),
        # reserved #
        ("user_pick_up", pgettext_lazy("Event type", "User pick up")),
        # unavailable #
        ("maintenance", pgettext_lazy("Event type", "Maintenance")),
        ("low_battery", pgettext_lazy("Event type", "Low battery")),
        # removed #
        ("service_end", pgettext_lazy("Event type", "Service end")),
        ("rebalance_pick_up", pgettext_lazy("Event type", "Rebalance pick up")),
        ("maintenance_pick_up", pgettext_lazy("Event type", "Maintenance pick up")),
        ("agency_pick_up", pgettext_lazy("Event type", "Agency pick up")),
        # this last event is in neither of the MDS spec
        ("telemetry", pgettext_lazy("Event type", "Received telemetry")),
        # This may be added in a revision of the agency API specs
        ("battery_charged", pgettext_lazy("Event type", "Battery charged")),
    ],
)

# see: https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/agency
EVENT_TYPE_REASON = enum.Enum(
    "Event type reason",
    [
        ("low_battery", pgettext_lazy("Event type reason", "Low battery")),
        ("maintenance", pgettext_lazy("Event type reason", "Maintenance")),
        ("compliance", pgettext_lazy("Event type reason", "Compliance")),
        ("off_hours", pgettext_lazy("Event type reason", "Off hours")),
        ("rebalance", pgettext_lazy("Event type reason", "Rebalance")),
        ("charge", pgettext_lazy("Event type reason", "Charge")),
        ("missing", pgettext_lazy("Event type reason", "Missing")),
        ("decommissioned", pgettext_lazy("Event type reason", "Decommissioned")),
    ],
)

EVENT_SOURCE = enum.Enum(
    "Event source",
    [
        ("agency_api", pgettext_lazy("Event source", "Agency API")),
        ("provider_api", pgettext_lazy("Event source", "Provider API")),
    ],
)

AREA_TYPE = enum.Enum(
    "Area type",
    [
        ("unrestricted", pgettext_lazy("Area type", "Unrestricted")),
        ("restricted", pgettext_lazy("Area type", "Restricted")),
        ("preferred_pick_up", pgettext_lazy("Area type", "Preferred pick up")),
        ("preferred_drop_off", pgettext_lazy("Area type", "Preferred drop off")),
    ],
)

EVENT_TYPE_TO_DEVICE_STATUS = {
    EVENT_TYPE.register.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.service_start.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.service_end.name: DEVICE_STATUS.unavailable.name,
    EVENT_TYPE.provider_drop_off.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.provider_pick_up.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.city_pick_up.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.reserve.name: DEVICE_STATUS.reserved.name,
    EVENT_TYPE.cancel_reservation.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.trip_start.name: DEVICE_STATUS.trip.name,
    EVENT_TYPE.trip_enter.name: DEVICE_STATUS.trip.name,
    EVENT_TYPE.trip_leave.name: DEVICE_STATUS.elsewhere.name,
    EVENT_TYPE.trip_end.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.deregister.name: DEVICE_STATUS.inactive.name,
    # Other events
    EVENT_TYPE.telemetry.name: DEVICE_STATUS.unknown.name,
    EVENT_TYPE.battery_charged.name: DEVICE_STATUS.available.name,
    # Old events
    OLD_EVENT_TYPE.rebalance_drop_off.name: DEVICE_STATUS.available.name,
    OLD_EVENT_TYPE.maintenance_drop_off.name: DEVICE_STATUS.available.name,
    OLD_EVENT_TYPE.low_battery.name: DEVICE_STATUS.unavailable.name,
    OLD_EVENT_TYPE.maintenance.name: DEVICE_STATUS.unavailable.name,
    OLD_EVENT_TYPE.rebalance_pick_up.name: DEVICE_STATUS.removed.name,
    OLD_EVENT_TYPE.maintenance_pick_up.name: DEVICE_STATUS.removed.name,
}



# Generated by Django 2.2 on 2019-09-23 11:16

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0045_add_fixed_price_on_policy_model'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='compliance',
            unique_together={('policy', 'rule', 'geography', 'vehicle', 'start_date')},
        ),
    ]
